/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#include "VXS_Bond.h"
#include "VXS_Voxel.h"
#include "VX_Sim.h"
//#include "RelaxMaterial.h"

#define PI 3.1415926

CVXS_Bond::CVXS_Bond(void)
{
	p_Sim = NULL;

	Vox1SInd = -1;
	Vox2SInd = -1;

//	Vox1 = NULL;
//	Vox2 = NULL;
	ThisBondType = B_LINEAR;
	Force1 = Vec3D(0,0,0);
	Force2 = Vec3D(0,0,0);
	Moment1 = Vec3D(0,0,0);
	Moment2 = Vec3D(0,0,0);
	
	Perm = false;

//	CurPVec = Vec3D(0,0,0);

//	CurStiffness = 0;
//	Compression = 1.0;
//	CurStress = 0.0;

	Yielded = false;
	Broken = false;


	L = Vec3D(0, 0, 0); E=0; u=0; G=0; A=0; Iy=0; Iz=0; J=0; a1=0; a2=0; b1y=0; b2y=0; b3y=0; b1z=0; b2z=0; b3z=0; CTE=0; //ini to zero!
	ElRot = CQuat(1,0,0,0);

	CurStrain = 0; CurStress = 0; MaxStrain = 0;

	_pV1Cache = NULL;
	_pV2Cache = NULL;
	_v1SIndCache = -1;
	_v2SIndCache = -1;


//	BondVel = Vec3D(0,0,0);
//	BondAngVel = Vec3D(0,0,0);
}

CVXS_Bond::~CVXS_Bond(void)
{
}

CVXS_Bond& CVXS_Bond::operator=(const CVXS_Bond& Bond)
{
	p_Sim = Bond.p_Sim;

	Vox1SInd = Bond.Vox1SInd;
	Vox2SInd = Bond.Vox2SInd;
	ThisBondType = Bond.ThisBondType;
//	Yielded = Bond.Yielded;
//	Broken = Bond.Broken;
//	CurStiffness = Bond.CurStiffness;
	Force1 = Bond.Force1;
	Force2 = Bond.Force2;

	Moment1 = Bond.Moment1;
	Moment2 = Bond.Moment2;

	RestDist = Bond.RestDist;
	OrigDist = Bond.OrigDist;

	Yielded = Bond.Yielded;
	Broken = Bond.Broken;


	L = Bond.L;
	E = Bond.E;
	u = Bond.u;
	G = Bond.G;
	A = Bond.A;
	Iy = Bond.Iy;
	Iz = Bond.Iz;
	J = Bond.J;
	a1 = Bond.a1;
	a2 = Bond.a2;
	b1y = Bond.b1y;
	b2y = Bond.b2y;
	b3y = Bond.b3y;
	b1z = Bond.b1z;
	b2z = Bond.b2z;
	b3z = Bond.b3z;
	CTE = Bond.CTE;
	ElRot = Bond.ElRot;
	CurStrain = Bond.CurStrain;
	CurStress = Bond.CurStress;
	MaxStrain = Bond.MaxStrain;
	Perm = Bond.Perm;

	_pV1Cache = Bond._pV1Cache;
	_pV2Cache = Bond._pV2Cache;
	_v1SIndCache = Bond._v1SIndCache;
	_v2SIndCache = Bond._v2SIndCache;

	return *this;
}

CVXS_Voxel* CVXS_Bond::Vox1(void)
{
	if (_v1SIndCache != Vox1SInd || _pV1Cache == NULL){
		_pV1Cache = &(p_Sim->VoxArray[Vox1SInd]); 
		_v1SIndCache = Vox1SInd;
	}
	return _pV1Cache;
}

CVXS_Voxel* CVXS_Bond::Vox2(void)
{
	if (_v2SIndCache != Vox2SInd || _pV2Cache == NULL){
		_pV2Cache = &(p_Sim->VoxArray[Vox2SInd]); 
		_v2SIndCache = Vox2SInd;
	}
	return _pV2Cache;
}

void CVXS_Bond::CalcForce() //calculates force, positive for tension, negative for compression
{

//	if (!Broken){
		switch (ThisBondType){
		case B_LINEAR: CalcLinForce(); break;
		case B_LINEAR_CONTACT: CalcContactForce(); break;
		case B_TILE_EDGE: CalcTileLookupForce(); break;
		case B_TILE_ZCORNER: CalcTileLookupForce(); break;
		case B_INPUT_LINEAR_NOROT: if (p_Sim->Dragging) CalcSimpleSpringForce(); else {Force1 = Vec3D(0,0,0); Force2 = Vec3D(0,0,0); Moment1 = Vec3D(0,0,0); Moment2 = Vec3D(0,0,0);} break;
		default: {}//AfxMessageBox("Unknown Bond type!");
		}

//	}
	//else {Force stays zero if broken...}

}


bool CVXS_Bond::UpdateBondParams(vfloat CurStrainIn)
{
//	CVXC_Material* pMat1 = p_Sim->LocalVXC.GetBaseMat(Vox1()->MatIndex); //for convenience. We'll reference these a few times...
//	CVXC_Material* pMat2 = p_Sim->LocalVXC.GetBaseMat(Vox2()->MatIndex);

//	if (!Vox1()->IsBlended && !Vox2()->IsBlended && pMat1->GetMatModel() == MDL_LINEAR && pMat2->GetMatModel() == MDL_LINEAR){ CurStress = E*CurStrainIn; return true;} //quick and easy catch for maximum speed linear simulation!


	vfloat Strain1, Strain2, Stress1, Stress2;
	bool Yielded1, Yielded2, Broken1, Broken2;

	if (!p_Sim->IsPlasticityEnabled() || CurStrainIn >= MaxStrain){ //if we're in new territory on the stress-strain curve or plasticity is not enabled...
		MaxStrain = CurStrainIn; //set the high-water mark on the strains

		Strain1 = CurStrainIn; //initial guesses at strains
		Strain2 = CurStrainIn;

		//get stiffness of each voxel in question...
		Vox1()->CalcVoxMatStress(Strain1, &Stress1, &Yielded1, &Broken1);
		Vox2()->CalcVoxMatStress(Strain2, &Stress2, &Yielded2, &Broken2);

		int MaxCount = 3;
		int count = 0;
		while (abs(Stress1-Stress2) > (Stress1+Stress2)/2e3 && count<MaxCount){ //refine guesses at the strains until difference is less than .1% of average stress
			Strain1 = 2*Stress2/(Stress1+Stress2)*Strain1;
			Strain2 = 2*Stress1/(Stress1+Stress2)*Strain2;

			Vox1()->CalcVoxMatStress(Strain1, &Stress1, &Yielded1, &Broken1);
			Vox2()->CalcVoxMatStress(Strain2, &Stress2, &Yielded2, &Broken2);
			count++;
		}

		CurStress = (Stress1+Stress2)/2; //average just in case we didn't quite converge to the exact same stress
			
		if (!Yielded && (Yielded1 || Yielded2)) SetYielded(); //if any part of the bond yielded, its all yielded
		if (!Broken && (Broken1 || Broken2)) SetBroken(); //if any part of the bond broke, its all broke

		//Update the rest distance for the voxels...
		if (p_Sim->IsPlasticityEnabled())
			RestDist = OrigDist*(1+(MaxStrain - CurStress/E));
	}
	else { //if we've backed off a max strain and linearly drop back down at the elastic modulus
		CurStress = E*(CurStrainIn-(RestDist.Length()/OrigDist.Length()-1.0));
	}

	return true;
}

void CVXS_Bond::SetYielded(void)
{
	Yielded = true;
	Vox1()->VYielded = true;
	Vox2()->VYielded = true;
}

void CVXS_Bond::SetBroken(void)
{
	if (p_Sim->IsFailureEnabled()){
		Broken = true; 
		Vox1()->VBroken = true;
		Vox2()->VBroken = true;
	}
}


bool CVXS_Bond::CalcConsts(void) //fills in the constant parameters for the bond...
{
	//store a rotation that goes from this bond angle to +X direction
//	Vec3D RetAngle;
//	double Angle = OrigDist.AlignWith(Vec3D(1,0,0), RetAngle);
//	ElRot = CQuat(Angle, RetAngle);
	ElRot = CQuat(OrigDist, Vec3D(1,0,0));

	vfloat E1 = Vox1()->GetEMod(); //p_Sim->LocalVXC.GetBaseMat(Vox1()->MatIndex)->GetElasticMod();
	vfloat E2 = Vox2()->GetEMod(); //p_Sim->LocalVXC.GetBaseMat(Vox2()->MatIndex)->GetElasticMod();
	vfloat u1 = Vox1()->GetPoisson(); //p_Sim->LocalVXC.GetBaseMat(Vox1()->MatIndex)->GetPoissonsRatio(); //poissons ratio of the two voxels in question...
	vfloat u2 = Vox2()->GetPoisson(); //p_Sim->LocalVXC.GetBaseMat(Vox2()->MatIndex)->GetPoissonsRatio();
	vfloat CTE1 = Vox1()->GetCTE(); //p_Sim->LocalVXC.GetBaseMat(Vox1()->MatIndex)->GetCTE(); //CTE of the two voxels in question...
	vfloat CTE2 = Vox2()->GetCTE(); //p_Sim->LocalVXC.GetBaseMat(Vox2()->MatIndex)->GetCTE();

	if (u1 < 0 || u1 > 0.5 || u2 < 0 || u2 > 0.5 ) {return false;} //AfxMessageBox("bad poissons ratio");

	//for now we are only using the nominal size of the voxel, although we could change this later if needed
	//rotate the box dimensions into the correct refference frame of x being direction of bond
	switch (ThisBondType){
		case B_LINEAR:
		L = ElRot.Conjugate()*CQuat((Vox1()->NomScale + Vox2()->NomScale)/2)*ElRot; //in X direction, so L.X is beam length, L.y, L.z are transverse dimensions
		L = L.Abs(); //distances strictly positive!
		break;
		case B_LINEAR_CONTACT: //X direction of L is in line with the contact...
		default:		
			L.x = p_Sim->LocalVXC.GetLatticeDim();
			L.y = L.x;
			L.z = L.x;
		break;
	}

//	CTE = (CTE1/2+CTE2/2); //thermal expansion
//	L.x *= 1 + CTE*(p_Sim->pEnv->CurTemp - p_Sim->pEnv->TempBase); //FOR ROTATE DEBUG...

	A = L.y * L.z;
	Iy = L.z*L.y*L.y*L.y / 12; //BHHH/12
	Iz = L.y*L.z*L.z*L.z / 12;
	J = L.y*L.z*(L.y*L.y + L.z*L.z)/12; //torsional MOI: BH/12*(BB+HH)

	if (E1 == 0 || E2 == 0) {return false;}

	E = (E1*E2/(E1+E2))*2; //x2 derived from case of equal stiffness: E1*E1/(E1+E1) = 0.5*E1 
	u = (u1*u2/(u1+u2))*2; //Poissons ratio
	CTE = (CTE1/2+CTE2/2); //thermal expansion
	G = E/(2*(1+u)); //Shear modulus is related to
	a1 = E * A / L.x;
	a2 = G * J / L.x;
	b1y = 12 * E * Iy / (L.x*L.x*L.x); // + G * A / L;
	b1z = 12 * E * Iz / (L.x*L.x*L.x); // + G * A / L;
	b2y = 6 * E * Iy / (L.x*L.x);
	b2z = 6 * E * Iz / (L.x*L.x);
	b3y = 2 * E * Iy / L.x;
	b3z = 2 * E * Iz / L.x;


	if (!UpdateBondParams(0)) return false;

	return true;
}

//sub force calculation types...
void CVXS_Bond::CalcLinForce() //get bond forces given positions, angles, and stiffnesses...
{

	//quaternion properties (for my reference)
	//1) To rotate a vector V, form a quaternion with w = 0; To rotate by Quaternion Q, do Q*V*Q.Conjugate() and trim off the w component.
	//2) To do multiple rotations: To Rotate by Q1 THEN Q2, Q2*Q1*V*Q1.Conjugate*Q2.Conjugate(), or make a Qtot = Q2*Q1 and do Qtot*V*Qtot.Conjugate()
	//3) Q1*Q1.Conjugate - Identity
	//4) To do a reverse rotation Q1, just do Q1.conjugate*V*Q1
	//5) An orientation quaternion is really just the relative rotation from the identity quaternion (1,0,0,0) to this orientation.
	//6) If an orientation is represented by Q1, to rotate that orientation by Q2 the new orientation is Q2*Q1
	//http://www.cprogramming.com/tutorial/3d/quaternions.html

		//local variables...
//	Vec3D RotAngle, Pos2, Vel2, Angle2;
	enum SimType {SIM_LINEAR_ALIGN_P1_A1, SIM_LINEAR_ALIGN_P1_P2};
//	SimType ThisSim = SIM_LINEAR_ALIGN_P1_A1;
	SimType ThisSim = SIM_LINEAR_ALIGN_P1_P2;

	switch (ThisSim){
	case SIM_LINEAR_ALIGN_P1_P2:{
		Vec3D CurRelPos = Vox2()->S.Pos - Vox1()->S.Pos; //current relative position between voxels

		//undo current voxel rotation to put in line with original bond according to Angle 1
		CQuat AlignedPos2 = Vox1()->S.Angle.Conjugate()*CurRelPos*Vox1()->S.Angle;

		//Rotate it so that Pos2 is now right in line with X (Angle 1 is now non-zero)
		CQuat AlignedToRestRot = CQuat(AlignedPos2.ToVec(), OrigDist); //axis rotating form Current to original direction

		//undo original bond orientation to put this in line with the X Xaxis
		Vec3D Pos2 = Vec3D(CurRelPos.Length() - (Vox1()->S.Scale.x + Vox2()->S.Scale.x)/2, 0, 0); //- TempFactor*L.x, 0, 0); //ElRot*Pos2Q*ElRot.Conjugate(); //Position of vox2
		CQuat Vox1Angle = ElRot * AlignedToRestRot * ElRot.Conjugate(); //AlignedToRestRot is the angle of Vox1 because it was aligned with OrigDist, then we rotated the element by AlignedToRestRot. These Elrot.Conjugate() at the end are because we are changing coordinate systems (I think...)
		CQuat Vox2Angle = ElRot * AlignedToRestRot * Vox1()->S.Angle.Conjugate() * Vox2()->S.Angle * ElRot.Conjugate(); //angle of Vox2
		CQuat TotalRot = ElRot * AlignedToRestRot * Vox1()->S.Angle.Conjugate();

		//render to rotation vector for FEM
		vfloat tmpRot;
		Vec3D Angle1;
		Vox1Angle.AngleAxis(&tmpRot, &Angle1.x, &Angle1.y, &Angle1.z); //direction cosines?
		Angle1 *= tmpRot;

		Vec3D Angle2;
		Vox2Angle.AngleAxis(&tmpRot, &Angle2.x, &Angle2.y, &Angle2.z); //direction cosines?
		Angle2 *= tmpRot;

		CurStrain = Pos2.x/L.x; //   abs(AlignedPos.ToVec().Length()/(Vox2()->NomPos-Vox1()->NomPos).Length()); //quick hack...
		UpdateBondParams(CurStrain); //updates the bond parameters (yileded, broken, stress...) based on the current Strain

		//simple beam bending equations, but since we've cleverly aligned, Pos1 = (0,0,0) and Pos2 = (X,0,0). So they drop out (except Pos2.x!)
		Force1.x = -CurStress*L.y*L.z; //Do it this way to account for non-linear deformation instead of -a1*Pos2.x
		Force1.y = b2z*Angle1.z + b2z*Angle2.z;
		Force1.z = -b2y*Angle1.y - b2y*Angle2.y;
		Moment1.x = a2*Angle1.x - a2*Angle2.x;
		Moment1.y = b2y*Pos2.z + 2*b3y*Angle1.y + b3y*Angle2.y;
		Moment1.z = -b2z*Pos2.y + 2*b3z*Angle1.z + b3z*Angle2.z;
		Moment2.x = -a2*Angle1.x + a2*Angle2.x;
		Moment2.y = b3y*Angle1.y + 2*b3y*Angle2.y;
		Moment2.z = b3z*Angle1.z + 2*b3z*Angle2.z;

		//unrotate
		CQuat InvRotate = TotalRot.Conjugate();
		Force1 = TotalRot.Conjugate()*CQuat(Force1)*TotalRot;
		Force2 = -Force1;
		Moment1 = TotalRot.Conjugate()*CQuat(Moment1)*TotalRot;
		Moment2 = TotalRot.Conjugate()*CQuat(Moment2)*TotalRot;
								}
		break;
	case SIM_LINEAR_ALIGN_P1_A1:{ //OLD!!!

		//local variables...
	//	Vec3D RotAngle, Pos2, Vel2, Angle2;
		vfloat tmpRot;


		//CalcConsts(); //TEMP, for rotate debug

		Vec3D RestBond = OrigDist; //add in nominal offsets, temperature changes, etc. here!
	//TODO: to  = RestDist for correct nonlinear behavior?!

		if (p_Sim->PoissonKickBackEnabled){
	///		Vec3D CurSize = ((Vox1()->S.Scale - Vox1()->NomScale) + (Vox2()->S.Scale - Vox1()->NomScale))/2; //how swollen/contracted in each dim we are
	///		double Length = CurSize.Dot(RestDist.Normalized());
	///		RestBond += RestBond/RestBond.Length()*Length;
		}
		if (p_Sim->pEnv->TempEnabled){	
		double CTE1 = p_Sim->LocalVXC.GetBaseMat(Vox1()->GetMatIndex())->GetCTE(); //CTE of the two voxels in question...
		double CTE2 = p_Sim->LocalVXC.GetBaseMat(Vox2()->GetMatIndex())->GetCTE();
		//double tempDelta = p_Sim->pEnv->CurTemp - p_Sim->pEnv->TempBase;
		double Expan1 = CTE1 * (p_Sim->LocalVXC.GetBaseMat(Vox1()->GetMatIndex())->GetMatTemp() - p_Sim->pEnv->TempBase);  //all temps are stored as absolute, and the activation is assumed to be relative to the Base Temperature
		double Expan2 = CTE2 * (p_Sim->LocalVXC.GetBaseMat(Vox2()->GetMatIndex())->GetMatTemp() - p_Sim->pEnv->TempBase);
		RestBond *= 1 + ((Expan1/2)+(Expan2/2)); //account for thermal expansion. Combine the thermal expansion of each Voxel
		}	

		//Rotate element by the inverse of angle 1 to line it up with the original bond axis (rotation of Vox1 = 0)
		CQuat CurRelPos = CQuat(Vox2()->S.Pos - Vox1()->S.Pos); //current relative position between voxels

		//undo current voxel rotation to put in line with original bond
		CQuat AlignedPos = Vox1()->S.Angle.Conjugate()*CurRelPos*Vox1()->S.Angle;
		CQuat AlignedAngle = Vox1()->S.Angle.Conjugate()*Vox2()->S.Angle; //rotate the vox-2 angle by Vox1
	//	AlignedAngle.Normalize(); //????????????????????????????

		//subtract nominal bond length, other offsets here... (to get change in position
		AlignedPos -= CQuat(RestBond);

		//undo original bond orientation to put this in line with the X Xaxis
		Vec3D Pos2 = ElRot*AlignedPos*ElRot.Conjugate(); //Position of vox2
		CQuat Vox2Angle = ElRot*AlignedAngle*ElRot.Conjugate(); //angle of Vox2

		//render to rotation vector for FEM
		Vec3D Angle2;
	//	Vox2Angle.Normalize(); ///?????????????????????????
		Vox2Angle.AngleAxis(tmpRot, Angle2);
		Angle2 *= tmpRot;


		CurStrain = Pos2.x/L.x; //   abs(AlignedPos.ToVec().Length()/(Vox2()->NomPos-Vox1()->NomPos).Length()); //quick hack...
		//L.x? possibly account for temperature here...

		UpdateBondParams(CurStrain); //updates the bond parameters (yileded, broken, stress...) based on the current Strain

	//TODO: deal with when, say, x is far extended, angles go wonky.

		//simple beam bending equations, but since we've cleverly rotated, Pos1 = 0,0,0 and Angle1 = (0,0,0). (So they drop out!) (Not sure can do this...)
		Force1.x = -a1*Pos2.x;
	///	Force1.x = -CurStress*L.y*L.z; //Do it this way to account for non-linear deformation
		Force1.y = -b1z*Pos2.y + b2z*Angle2.z;
		Force1.z = -b1y*Pos2.z - b2y*Angle2.y;
		Moment1.x = -a2*Angle2.x;
		Moment1.y = b2y*Pos2.z + b3y*Angle2.y;
		Moment1.z = -b2z*Pos2.y + b3z*Angle2.z;
	///	Moment1.y = b2y*Pos2.z/(1+Pos2.x/L.x) + b3y*Angle2.y;
	///	Moment1.z = -b2z*Pos2.y/(1+Pos2.x/L.x) + b3z*Angle2.z;
		Force2 = -Force1;
		Moment2.x = a2*Angle2.x;
	///	Moment2.y = b2y*Pos2.z/(1+Pos2.x/L.x) + 2*b3y*Angle2.y; //these 2's dont seem right... Moments should be the same both ways! but equations don't lie...
	///	Moment2.z = -b2z*Pos2.y/(1+Pos2.x/L.x) + 2*b3z*Angle2.z;
		Moment2.y = b2y*Pos2.z + 2*b3y*Angle2.y; //these 2's dont seem right... Moments should be the same both ways! but equations don't lie...
		Moment2.z = -b2z*Pos2.y + 2*b3z*Angle2.z;


		//unrotate
		CQuat InvRotate = Vox1()->S.Angle*ElRot.Conjugate();
		Force1 = InvRotate*CQuat(Force1)*InvRotate.Conjugate();
		Force2 = InvRotate*CQuat(Force2)*InvRotate.Conjugate();
		Moment1 = InvRotate*CQuat(Moment1)*InvRotate.Conjugate();
		Moment2 = InvRotate*CQuat(Moment2)*InvRotate.Conjugate();
								}
		break;
	}







	//Damping: (in world coordinate system now!
	Vec3D avgBondVel = (Vox2()->S.Vel + Vox1()->S.Vel)/2; 
	Vec3D avgBondAngVel = (Vox2()->S.AngVel + Vox1()->S.AngVel)/2; 
	Vec3D avgBondPos = (Vox2()->S.Pos + Vox1()->S.Pos)/2;
		
	//linear
	Vec3D RotComp = (Vox2()->S.Pos-avgBondPos).Cross(avgBondAngVel);
//	Vec3D BondVel = Vox1()->S.Vel - (avgBondVel - RotComp); //if this, switch signs below!
	Vec3D BondVel = Vox2()->S.Vel - avgBondVel + RotComp;

	Force1 -= p_Sim->GetBondDampZ()*2*sqrt(a1*Vox1()->Mass)*BondVel; //Damping force for this bond: Fdamping = c*Vel, c = zeta*2*Mass*w0 = zeta*2*sqrt(mk) //http://en.wikipedia.org/wiki/Damping
	Force2 += p_Sim->GetBondDampZ()*2*sqrt(a1*Vox2()->Mass)*BondVel; //Damping force for this bond: Fdamping = c*Vel, c = zeta*2*Mass*w0 //http://en.wikipedia.org/wiki/Damping

	
	//rotational
//	double zeta = 1.0;
	Vec3D BondAngVel = Vox2()->S.AngVel - Vox1()->S.AngVel;
	Moment1 -= p_Sim->GetBondDampZ()*2*sqrt(a2*Vox1()->Inertia)*BondAngVel/2; //not sure about these twos!!
	Moment2 += p_Sim->GetBondDampZ()*2*sqrt(a2*Vox2()->Inertia)*BondAngVel/2;

//	Moment1 -= zeta*2*sqrt(a2*Vox1()->Inertia)*BondAngVel/2; //not sure about these twos!!
//	Moment2 += zeta*2*sqrt(a2*Vox2()->Inertia)*BondAngVel/2;

}

void CVXS_Bond::CalcContactForce() 
{
	//just basic sphere envelope, repel with the stiffness of the material... (assumes CalcConsts has been called)

	Vec3D Pos2 = Vox2()->S.Pos - Vox1()->S.Pos;
	vfloat NomDist = (Vox1()->S.Scale.x/2 + Vox2()->S.Scale.x/2)*1.2; //effective diameter of 1.5 voxels...
	vfloat RelDist = NomDist - Pos2.Length(); //positive for overlap!
	if (RelDist > 0){ //if we're overlapping
		Force1 = Pos2/Pos2.Length() *a1*(RelDist); // + 2*RelDist*RelDist/L); //unit vector in direction of bond, plus linear term, plue squared term
		Force2 = -Force1; // + 2*RelDist*RelDist/L); //unit vector in direction of bond, plus linear term, plue squared term

	}
	else {
		Force1 = Vec3D(0,0,0);
		Force2 = Vec3D(0,0,0);

	}

	Moment1 = Vec3D(0,0,0);
	Moment2 = Vec3D(0,0,0);

}

void CVXS_Bond::CalcSimpleSpringForce()
{
	Vec3D CDist = Vox2()->S.Pos - Vox1()->S.Pos;

	Force1 = a1*CDist;
	Force2 = -a1*CDist;
	Moment1 = Vec3D(0,0,0);
	Moment2 = Vec3D(0,0,0);
}

Vec3D CVXS_Bond::CalcTileLookupForce(void)
{
	/*
	Vec3D ThisForce = Vec3D(0,0,0);
	CurPVec = Vec3D(0,0,0); //dump torques...

	Vec3D tmpForce = Vec3D(0,0,0);
	Vec3D tmpForce2 = Vec3D(0,0,0);
	Vec3D tmpTorque = Vec3D(0,0,0);
	Vec3D tmpTorque2 = Vec3D(0,0,0);
	double tmpStress1 = 0;
	double tmpStress2 = 0;
	Vec3D LP1 = Vec3D(0,0,0);
	Vec3D LP2 = Vec3D(0,0,0);
	bool WantFlipX = false;
	bool WantFlipY = false;
	bool WantXDir = false;
	double LocalAngle = 0.0;

	//the positions of the voxels...
	Vec3D P1 = Vox1()->Pos;
	Vec3D P2 = Vox2()->Pos;
	Vec3D PVec = P2-P1;

	Vec3D SlackVec;

	double dTemp = 0; //p_Sim->CurTemp - p_Sim->RefTemp;

	double Dist = PVec.Length();

	double P1x;
	double P1y;
	double P2x;
	double P2y;

	//Here, account for interaction point of square tiles!
	switch (ThisBondType){
	case B_TILE_EDGE:
		//Only two possibilities, because we took care in ordering things

		if (abs(PVec.x) > abs(PVec.y)) { //if in the +X direction:
			WantXDir = true;
			P1.x = Vox1()->Pos.x + Vox1()->ActSize.x/2*(1 + p_Sim->pMaterials[Vox1()->MatIndex].CTE * dTemp);
			P2.x = Vox2()->Pos.x - Vox2()->ActSize.x/2*(1 + p_Sim->pMaterials[Vox2()->MatIndex].CTE * dTemp);
		}
		else if (abs(PVec.y) > abs(PVec.x)) { //if in the +Y direction:
			WantXDir = false;
			P1.y = Vox1()->Pos.y + Vox1()->ActSize.y/2*(1 + p_Sim->pMaterials[Vox1()->MatIndex].CTE * dTemp);
			P2.y = Vox2()->Pos.y - Vox2()->ActSize.y/2*(1 + p_Sim->pMaterials[Vox2()->MatIndex].CTE * dTemp);
		}

		PVec = (P2-P1)/1000; //convert to mm from meters

		//calculate forces and torques!!
		if (Vox1()->MatIndex == Vox2()->MatIndex){ //If the same material:
			//this function takes PVec input as mm!
			CurStress = (*(p_Sim->pMaterials + Vox1()->MatIndex)).GetForce(B_TILE_EDGE, &ThisForce, &PVec, dTemp, WantFlipX, WantFlipY, WantXDir);
		}

		else { //if we need to combine two materials
			tmpStress1 = p_Sim->pMaterials[Vox1()->MatIndex].GetForce(B_TILE_EDGE, &tmpForce, &PVec, dTemp, WantFlipX, WantFlipY, WantXDir);
			tmpStress2 = p_Sim->pMaterials[Vox2()->MatIndex].GetForce(B_TILE_EDGE, &tmpForce2, &PVec, dTemp, WantFlipX, WantFlipY, WantXDir);

			ThisForce.x = 2/(1/tmpForce.x + 1/tmpForce2.x); //First order approximation of interaction of different modulus materials
			ThisForce.y = 2/(1/tmpForce.y + 1/tmpForce2.y); //First order approximation of interaction of different modulus 
			//Torque.z = 2/(1/tmpTorque.y + 1/tmpTorque2.y);
			CurStress = 2/(1/tmpStress1 + 1/tmpStress2);
		}
		
		ThisForce.z = 0; //Separate calculation here....

		CurPVec = PVec;
		break;

	case B_TILE_ZCORNER:
		LocalAngle = 0.0;
		//set flips and angles: (to take care of the 4 different bond orientations!
		if (PVec.x > 0) {
			WantFlipX = false;	
			if (PVec.y > 0) LocalAngle = PI/4;
			else LocalAngle = 7*PI/4;
		}
		else { //PVec.x < 0
			WantFlipX = true;
			if(PVec.y > 0) LocalAngle = 3*PI/4;
			else LocalAngle = 5*PI/4;
		}

		if (PVec.y > 0) WantFlipY = false;
		else {
			WantFlipY = true;
		}
		//rendered quadrant positions (approximation to theta)
		P1x = P1.x + cos(LocalAngle)*(Vox1()->ActSize.x*sqrt(2.0f)/4) * (1 + (*(p_Sim->pMaterials + Vox1()->MatIndex)).CTE * dTemp);
		P1y = P1.y + sin(LocalAngle)*(Vox1()->ActSize.y*sqrt(2.0f)/4) * (1 + (*(p_Sim->pMaterials + Vox1()->MatIndex)).CTE * dTemp);
		P2x = P2.x + cos(LocalAngle + PI)*(Vox2()->ActSize.x*sqrt(2.0f)/4) * (1 + (*(p_Sim->pMaterials + Vox2()->MatIndex)).CTE * dTemp);
		P2y = P2.y + sin(LocalAngle + PI)*(Vox2()->ActSize.y*sqrt(2.0f)/4) * (1 + (*(p_Sim->pMaterials + Vox2()->MatIndex)).CTE * dTemp);

		LP1 = Vec3D(P1x, P1y, 0);
		LP2 = Vec3D(P2x, P2y, 0);

		PVec = (LP2-LP1)/1000; //conversion to mm (CS used to be 1=mm, now 1=m)

		//Get the X&Y force:
		if (Vox1()->MatIndex == Vox2()->MatIndex){ //If the same material:
			CurStress = (*(p_Sim->pMaterials + Vox1()->MatIndex)).GetForce(B_TILE_ZCORNER, &ThisForce, &PVec, dTemp, WantFlipX, WantFlipY); //XXX todo: normalize by area!!!!
		}
		else { //if we need to combine two materials
			tmpStress1 = (*(p_Sim->pMaterials + Vox1()->MatIndex)).GetForce(B_TILE_ZCORNER, &tmpForce, &PVec, dTemp, WantFlipX, WantFlipY);
			tmpStress2 = (*(p_Sim->pMaterials + Vox2()->MatIndex)).GetForce(B_TILE_ZCORNER, &tmpForce2, &PVec,  dTemp, WantFlipX, WantFlipY);

			ThisForce.x = 2/(1/tmpForce.x + 1/tmpForce2.x); //First order approximation of interaction of different mosulus materials
			ThisForce.y = 2/(1/tmpForce.y + 1/tmpForce2.y); //First order approximation of interaction of different mosulus 
			CurStress = 2/(1/tmpStress1 + 1/tmpStress2);

		}
		
		ThisForce = -ThisForce;
		ThisForce.z = 0; //Separate calculation here....

		CurPVec = PVec;
		break;

	default:
		AfxMessageBox("Unknown bond type");
	}

	return ThisForce;
	*/
	return Vec3D(0,0,0);
}
/*
void CVXS_Bond::UpdateStatus(void)
{

	double ThisYield = (*(p_Sim->pMaterials + Vox1()->MatIndex)).SigYield;
	double ThisUlt = (*(p_Sim->pMaterials + Vox1()->MatIndex)).SigUlt;

	if (ThisYield != 0 && CurStress > ThisYield){
		Yielded = true;
		Vox1()->VYielded = true;
		Vox2()->VYielded = true;
	}
	if (ThisUlt != 0 && CurStress > ThisUlt)
	{
		Broken = true;
		Vox1()->VBroken = true;
		Vox2()->VBroken = true;
	}

}

void CVXS_Bond::ResetStatus(void)
{
	Yielded = false;
	Broken = false;

}
*/
