/*******************************************************************************
Copyright (c) 2010, Jonathan Hiller (Cornell University)
If used in publication cite "J. Hiller and H. Lipson "Dynamic Simulation of Soft Heterogeneous Objects" In press. (2011)"

This file is part of Voxelyze.
Voxelyze is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
Voxelyze is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
See <http://www.opensource.org/licenses/lgpl-3.0.html> for license details.
*******************************************************************************/

#include "VXS_Voxel.h"
//#include "OpenGL.h"
#include "VXS_Bond.h"
#include "VX_Sim.h"

//#include "RelaxMaterial.h"

CVXS_Voxel::CVXS_Voxel(void)
{
	p_Sim = NULL;

	MatIndex = 0;

	m_Red = 0;
	m_Green = 0;
	m_Blue = 0;
	m_Trans = 1.0;

	BondInds.clear();

	NomScale = Vec3D(0,0,0);
	Mass = 1; //start with unit mass
	Inertia = 1; //needs to be calc'd
	ScaleMass = 1;

	CurMaxStrain = 0;
	CurMaxStress = 0;

	MyXIndex = -1;
	MySIndex = -1;

	Fixed = false;
	StaticFricFlag = false;
//	Displacable = false;
	NomPos = Vec3D(0,0,0); //nominal position, if this is fixed
	ExtForceApplied = Vec3D(0,0,0);
	ExtDisplaceApplied = Vec3D(0,0,0);
//	DispPos = Vec3D(0,0,0);
	VYielded = false;
	VBroken = false;

	FluidResistance = 0.0;

//	ExternalForce = Vec3D(0,0,0);
//	FrictionForce = Vec3D(0,0,0);

//	NumNearby = 0;

}

CVXS_Voxel::~CVXS_Voxel(void)
{
}

CVXS_Voxel& CVXS_Voxel::operator=(const CVXS_Voxel& VIn)
{
	p_Sim = VIn.p_Sim;
	S = VIn.S;
	PrevS = S;
	dS = VIn.dS;

	CurMaxStrain = VIn.CurMaxStrain;
	CurMaxStress = VIn.CurMaxStress;

	VYielded = VIn.VYielded;
	VBroken = VIn.VBroken;


	BondInds = VIn.BondInds;
	NearbyVoxInds = VIn.NearbyVoxInds;
	Fixed = VIn.Fixed;
	Mass = VIn.Mass;
	Inertia = VIn.Inertia;
	ScaleMass = VIn.ScaleMass;
	MyXIndex = VIn.MyXIndex;
	MySIndex = VIn.MySIndex;

	MatIndex = VIn.MatIndex;
	BlendMix = VIn.BlendMix;

	NomPos = VIn.NomPos;
	NomScale = VIn.NomScale;
	StaticFricFlag = VIn.StaticFricFlag;

	ExtForceApplied = VIn.ExtForceApplied;
	ExtDisplaceApplied = VIn.ExtDisplaceApplied;
	m_Red = VIn.m_Red;
	m_Green = VIn.m_Green;
	m_Blue = VIn.m_Blue;
	m_Trans = VIn.m_Trans;

	return *this;

}

CVXS_Bond* CVXS_Voxel::pBond(int LocalBondIndex)
{
	return &(p_Sim->BondArray[BondInds[LocalBondIndex]]);
}

CVXS_Voxel* CVXS_Voxel::pNearbyVox(int LocalNearbyInd)
{
	return &(p_Sim->VoxArray[NearbyVoxInds[LocalNearbyInd]]);
}

void CVXS_Voxel::UnlinkTmpBonds(void) //(assumes all perm at beginning...) removes any temporary bonds from our local list...
{
	while (BondInds.size() != 0 && !pBond(NumLocalBonds()-1)->Perm) BondInds.pop_back();
} 


void CVXS_Voxel::UpdateDState() //calculates the derivative of our current state "S". All bonds should be updates before this!
{
	dS.LinMom = S.LinMom;
	dS.AngMom = S.AngMom;
//	dS.ScaleMom = S.ScaleMom; //this isn't quite the right mass... need some sort of scaling inertia... Mass/6?

	dS.Force = CalcTotalForce();
	dS.Moment = CalcTotalMoment();
//	dS.ScaleForce = CalcTotalScaleForce();

}

void CVXS_Voxel::UpdateState(vfloat dtIn, dVoxState& Derivative) //advances this state "S" with previously calculated dS by dt
{
	/*
	//DISPLACEMENT
	S.LinMom = S.LinMom + Derivative.Force*dtIn;
//	S.Pos = S.Pos + dS.LinMom/Mass*dt; //dS.Vel = dS.LinMom/Mass
	S.Pos = S.Pos + Derivative.LinMom/Mass*dtIn; //dS.Vel = dS.LinMom/Mass

	//ANGLE
	S.AngMom = S.AngMom + Derivative.Moment*dtIn;

	//convert Angular velocity to quaternion form ("Spin")
//	Vec3D dSAngVel = dS.AngMom / Inertia;
	Vec3D dSAngVel = Derivative.AngMom / Inertia;
	if (abs(S.Angle.Length2()-1)>1e-6) S.Angle.Normalize(); //don't have to do this every time. Just if it drifts too much...
	CQuat Spin = 0.5 * CQuat(0, dSAngVel.x, dSAngVel.y, dSAngVel.z) * S.Angle; //current "angular velocity"


	S.Angle = S.Angle + Spin*dtIn; //see above

	//SCALE
//	S.ScaleMom = S.ScaleMom + Derivative.ScaleForce*dtIn;
//	S.Scale = S.Scale + dS.ScaleMom/ScaleMass*dtIn; //dS.ScaleVel = dS.ScaleMom/ScaleMass;
//	S.Scale = S.Scale + S.ScaleMom/ScaleMass*dtIn; //dS.ScaleVel = dS.ScaleMom/ScaleMass;

//temporary Temp update...
	//vfloat TempFact = 1.0;
	//if(p_Sim->pEnv->TempEnabled) TempFact = (1+(p_Sim->pEnv->CurTemp-p_Sim->pEnv->TempBase)*GetCTE()); //LocalVXC.GetBaseMat(VoxArray[i].MatIndex)->GetCTE());
	//if (TempFact < MIN_TEMP_FACTOR) TempFact = MIN_TEMP_FACTOR;
	//S.Scale = TempFact*NomScale;


	//Recalculate secondary:
	S.Vel = S.LinMom/Mass;
	S.AngVel = S.AngMom / Inertia;
//	S.ScaleVel = S.ScaleMom/ScaleMass;
*/
}

void CVXS_Voxel::EulerStep(void)
{
	TotVoxForce= CalcTotalForce(); //debug
	Vec3D ForceTot = TotVoxForce;
	//DISPLACEMENT
	S.LinMom = S.LinMom + ForceTot*p_Sim->dt;

	Vec3D Disp;
	if (Fixed) Disp = NomPos + ExtDisplaceApplied - S.Pos; //the exact diplacement needed to move it to its fixed position
	else Disp = S.LinMom/Mass*p_Sim->dt; //vector of what the voxel moves

	if(p_Sim->IsMaxVelLimitEnabled()){ //check to make sure we're not going over the speed limit!
		vfloat DispMag = Disp.Length();
		vfloat MaxDisp = p_Sim->GetMaxVoxVelLimit()*p_Sim->pEnv->pObj->GetLatticeDim();
		if (DispMag>MaxDisp) Disp *= (MaxDisp/DispMag);
	}
	S.Pos = S.Pos + Disp; //update position

	//ANGLE
	TotVoxMoment = CalcTotalMoment(); //debug
	S.AngMom = S.AngMom + TotVoxMoment*p_Sim->dt;

	//convert Angular velocity to quaternion form ("Spin")
	Vec3D dSAngVel = S.AngMom / Inertia;
	if (abs(S.Angle.Length2()-1)>1e-16) S.Angle.Normalize(); //don't have to do this every time. Just if it drifts too much...
	CQuat Spin = 0.5 * CQuat(0, dSAngVel.x, dSAngVel.y, dSAngVel.z) * S.Angle; //current "angular velocity"
	S.Angle = S.Angle + Spin*p_Sim->dt; //see above

	S.Angle.Normalize();

	//SCALE
//	S.ScaleMom = S.ScaleMom + CalcTotalScaleForce()*p_Sim->dt;
	vfloat TempFact = 1.0;
	if(p_Sim->pEnv->TempEnabled){ 
		//TempFact = (1+(p_Sim->pEnv->CurTemp-p_Sim->pEnv->TempBase)*GetCTE()); //LocalVXC.GetBaseMat(VoxArray[i].MatIndex)->GetCTE());
		TempFact = (1+(p_Sim->LocalVXC.GetBaseMat(GetMatIndex())->GetMatTemp() - p_Sim->pEnv->TempBase)*GetCTE());	//RBM Mod to allow selective temperature actuation for each different material
		}
	if (TempFact < MIN_TEMP_FACTOR) TempFact = MIN_TEMP_FACTOR;
	S.Scale = TempFact*NomScale;


	//Recalculate secondary:
	S.Vel = S.LinMom/Mass;
	S.AngVel = S.AngMom / Inertia;
//	S.ScaleVel = S.ScaleMom/ScaleMass;


	//tmp for plotting
	dS.Force = ForceTot;
}

void CVXS_Voxel::VerletStep(vfloat dt)
{
	Vec3D newPos = 2*S.Pos - PrevS.Pos + CalcTotalForce()/Mass*dt*dt;
	if (StaticFricFlag) {newPos.x = S.Pos.x; newPos.y = S.Pos.y;}

	PrevS.Pos = S.Pos;

	S.Vel = (newPos - S.Pos)/dt; //get Vel for other functions to use...
	S.Pos = newPos;

	if (Fixed) S.Pos = NomPos + ExtDisplaceApplied; //enforce fixed voxels...

}


void CVXS_Voxel::SetGLColor(void)
{
//	glColor4d(p_Sim->pMaterials[MatIndex].m_Red, p_Sim->pMaterials[MatIndex].m_Green, p_Sim->pMaterials[MatIndex].m_Blue, 1.0);

	//if(Fixed)
	//	glColor4d(0.0f, 0.5f, 1.0f, 1.0f);

	//else{ //if not fixed...
	//	float DVel = (float)Vel.Length()/NomScale.x/NomScale.x; ///1000;
	//	if (DVel > 1) DVel = 1;

	//	glColor4d(1.0, 1.0-DVel, 1.0-DVel, m_Trans);
	//}

}

void CVXS_Voxel::SetColor(float r, float g, float b, float a)
{
	m_Red = r;
	m_Green = g;
	m_Blue = b;
	m_Trans = a;
}	


Vec3D CVXS_Voxel::CalcTotalForce(bool WithRestraint)
{
	//INTERNAL forces
	Vec3D TotalForce = Vec3D(0,0,0);
	CurMaxStrain = 0;
	CurMaxStress = 0;

	//From the bonds
	int NumLocBond = NumLocalBonds();
	for (int i=0; i<NumLocBond; i++){
		if (IsMe(pBond(i)->Vox1())) TotalForce -= pBond(i)->Force1; //Force on Vox 1 from this bond
		else TotalForce -= pBond(i)->Force2; //Force on Vox 2 from this bond

		if (pBond(i)->ThisBondType == B_LINEAR){
			if (abs(pBond(i)->CurStrain) > CurMaxStrain) CurMaxStrain = abs(pBond(i)->CurStrain); //keep track of the most stress acting on this voxel...
			if (abs(pBond(i)->CurStress) > CurMaxStress) CurMaxStress = abs(pBond(i)->CurStress);
		}
	}

	//from interactive click/drag
	if (p_Sim->Dragging && MyXIndex == p_Sim->CurXSel){ //if we're currently dragging this one!
		TotalForce += p_Sim->InputBond()->Force1;
		TotalForce -= 2*Mass*sqrt(p_Sim->InputBond()->a1/Mass)*S.Vel;  //critically damp force for this bond to ground
	}

	//From gravity
	if (p_Sim->pEnv->GravEnabled)
		TotalForce.z += Mass*p_Sim->pEnv->GravAcc;

	//EXTERNAL forces
	TotalForce += CalcGNDDampEffect();
	if(p_Sim->FluidDampEnabled) CalcFluidEffect();
	TotalForce += ExtForceApplied; //add in any external forces....

	if(p_Sim->pEnv->FloorEnabled) TotalForce += CalcFloorEffect(TotalForce);
	else StaticFricFlag = false;

	if (StaticFricFlag) {TotalForce.x = 0; TotalForce.y = 0;}
	if (Fixed && WithRestraint) TotalForce = Vec3D(0,0,0);

//	TotalForce.y = 0;
//	TotalForce.z = 0;

	return TotalForce;
}

Vec3D CVXS_Voxel::CalcTotalMoment(void)
{
	Vec3D TotalMoment = Vec3D(0,0,0);
	for (int i=0; i<NumLocalBonds(); i++) {
		//add moments from bond
		if (MySIndex == pBond(i)->Vox1()->MySIndex){ TotalMoment -= pBond(i)->Moment1; } //if this is voxel 1
		else { TotalMoment -= pBond(i)->Moment2; } //if this is voxel 2
	}

	if (Fixed) TotalMoment = Vec3D(0,0,0);

//	TotalMoment.x = 0;
//	TotalMoment.y = 0;
//	TotalMoment.z = 0;


	return TotalMoment;
}

Vec3D CVXS_Voxel::CalcTotalScaleForce(void) //calculate Scaling force
{
	Vec3D TotalSForce = Vec3D(0,0,0);

	Vec3D TmpThisForce;

	for (int i=0; i<NumLocalBonds(); i++){
		TmpThisForce -= pBond(i)->Force1; //positive for tension, negative for compression
	}

	//add forces to enforce poissons ratio...
	vfloat MyStiff = p_Sim->LocalVXC.GetLeafMat(MatIndex)->GetElasticMod()*p_Sim->LocalVXC.GetLatticeDim();
	TotalSForce += (S.Angle.Conjugate()*CQuat(TmpThisForce)*S.Angle).ToVec(); //add in the force internal to the voxel...
	TotalSForce -= 2*MyStiff*(S.Scale-NomScale);
	vfloat poiss = 0.5;
	if (p_Sim->PoissonKickBackEnabled) {
		vfloat VolScale = NomScale.x*NomScale.y*NomScale.z/(S.Scale.x*S.Scale.y*S.Scale.z); //>1 if we're too small <1 if we're too large
		TotalSForce -= (2*poiss)*(1-VolScale) * MyStiff*(S.Scale); //arbitrary scaling factor...
	}

	TotalSForce -= 2*ScaleMass*sqrt(MyStiff/ScaleMass)*S.ScaleVel;  //Damping force for this bond:

	if (Fixed) TotalSForce = Vec3D(0,0,0);
	return TotalSForce;
}

Vec3D CVXS_Voxel::CalcFloorEffect(Vec3D TotalVoxForce) //calculates the object's interaction with a floor. should be calculated AFTER all other forces for static friction to work right...
{
	Vec3D FloorForce = Vec3D(0,0,0); //the force added by floor interactions...

	StaticFricFlag = false; //assume not under static friction unless we decide otherwise
	vfloat LatDim = (S.Scale.x+S.Scale.y+S.Scale.z)/3.0; //first order approximation!!! not accurate for oblong voxels!  //p_Sim->LocalVXC.GetLatticeDim(); //need to account for the change in volume elements...

	
	if (S.Pos.z < LatDim/2){ 
		vfloat NormalForce = p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetElasticMod() * LatDim/2 *(LatDim/2 - S.Pos.z); //positive for penetration...
		//TODO:																get id of this /2 ^  ?


		FloorForce.z += NormalForce; //force resisting penetration
		vfloat LocA1 = p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetElasticMod()*LatDim; //FIX THIS!!!!
		vfloat LocUDynamic = p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetuDynamic();
		vfloat LocUStatic = p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetuStatic();
	
		//do vertical damping here...
		FloorForce.z -= p_Sim->GetCollisionDampZ()*2*Mass*sqrt(LocA1/Mass)*S.Vel.z;  //critically damp force for this bond to ground


		vfloat SurfaceVel = sqrt(S.Vel.x*S.Vel.x + S.Vel.y*S.Vel.y); //velocity along the floor...
		vfloat SurfaceVelAngle = atan2(S.Vel.y, S.Vel.x); //angle of sliding along floor...
		vfloat SurfaceForce = sqrt(TotalVoxForce.x*TotalVoxForce.x + TotalVoxForce.y*TotalVoxForce.y);
		vfloat dFrictionForce = LocUDynamic*NormalForce; //TODO: make this a material property!
		Vec3D FricForceToAdd = -Vec3D(cos(SurfaceVelAngle)*dFrictionForce, sin(SurfaceVelAngle)*dFrictionForce, 0); //always acts in direction opposed to velocity in DYNAMIC friction mode
		//alwyas acts in direction opposite to force in STATIC friction mode

		if (S.Vel.x == 0 && S.Vel.y == 0){ //STATIC FRICTION: if this point is stopped and in the static friction mode...

			if (SurfaceForce < LocUStatic*NormalForce){ //if we don't have enough to break static friction
				StaticFricFlag = true;
			}
//			else //i don't think we need. this is when we've just broken static friction, so there's no velocity yet to have a counter force
//				CollisionForce += FricForceToAdd;

		}
		else { //DYNAMIC FRICTION
			if (dFrictionForce*p_Sim->dt < Mass*SurfaceVel){ // if we are sliding, coming to a stop. We know they are always in same angle, opposite directions.
				FloorForce += FricForceToAdd;
			}
			else {
				StaticFricFlag = true;
				S.LinMom.x = 0; //fully stop the voxel here! (caution...)
				S.LinMom.y = 0;
			}
		}
		
	}
	return FloorForce;

}

Vec3D CVXS_Voxel::CalcGNDDampEffect(void) //damps everything to ground as qucik as possible...
{
	Vec3D ToReturn = Vec3D(0,0,0);
	for (int i=0; i<NumLocalBonds(); i++){
		ToReturn -= p_Sim->GetSlowDampZ()*2*sqrt(pBond(i)->a1*Mass)*S.Vel;
	}

	return ToReturn;
}

void CVXS_Voxel::CalcFluidEffect(void)
{
	/*
	//add a force for each face normal to it (but no shear component...)
	//from vel..
	//first rotate the velocity into the voxel's original coordinate frame...
	Vec3D RotAngle, RotVel, RotForce;
	vfloat RotAmt;

	Angle.AngleAxis(RotAmt, RotAngle);
	RotVel = Vel; //Vel.Rot(RotAngle, -RotAmt); //velocity in the voxels original orientation

	bool oPX = true, oNX = true, oPY = true, oNY = true, oPZ = true, oNZ = true;

	//check the bonds to see where neighbors exist
	for (int i=0; i<m_NumPermBonds; i++){
		Vec3D oNomPos; //The position of the other bond
		float LatDim = pBond(i)->L; //lattice dimension

		if (this == pBond(i)->Vox1) oNomPos = pBond(i)->Vox2()->NomPos;
		else oNomPos = pBond(i)->Vox1()->NomPos;
		
		if (NomPos.y == oNomPos.y && NomPos.z == oNomPos.z){ //if Bond in X Axis...
			if (oNomPos.x > NomPos.x && oNomPos.x < NomPos.x + 1.5*LatDim) oPX = false;
			if (oNomPos.x < NomPos.x && oNomPos.x > NomPos.x - 1.5*LatDim) oNX = false;
		}
		if (NomPos.x == oNomPos.x && NomPos.z == oNomPos.z){ //if Bond in Y Axis...
			if (oNomPos.y > NomPos.y && oNomPos.y < NomPos.y + 1.5*LatDim) oPY = false;
			if (oNomPos.y < NomPos.y && oNomPos.y > NomPos.y - 1.5*LatDim) oNY = false;
		}
		if (NomPos.x == oNomPos.x && NomPos.y == oNomPos.y){ //if Bond in Z Axis...
			if (oNomPos.z > NomPos.z && oNomPos.z < NomPos.z + 1.5*LatDim) oPZ = false;
			if (oNomPos.z < NomPos.z && oNomPos.z > NomPos.z - 1.5*LatDim) oNZ = false;
		}
	}

	RotForce = Vec3D(0,0,0);
	vfloat Visc = 1; //viscosity
	Visc /= CurSize.x * CurSize.x; //normalize by area...
	if (oPX) RotForce.x -= Visc*Vel.x*abs(Vel.x);
	if (oNX) RotForce.x -= Visc*Vel.x*abs(Vel.x);
	if (oPY) RotForce.y -= Visc*Vel.y*abs(Vel.y);
	if (oNY) RotForce.y -= Visc*Vel.y*abs(Vel.y);
	if (oPZ) RotForce.z -= Visc*Vel.z*abs(Vel.z);
	if (oNZ) RotForce.z -= Visc*Vel.z*abs(Vel.z);

	Vec3D RealForce = RotForce; //RotForce.Rot(RotAngle, RotAmt);

	FluidResistance = abs(RealForce.Dot(Vel));
	if (FluidResistance < 1.0)
		int a = 0;

	if (FluidResistance > 1)
		int a = 1;

	return;
	//	ExternalForce += RealForce;
	//modify ExternalForce here...
	*/
}

/*
void CVXS_Voxel::ResetStatus(void)
{
	VYielded = false;
	VBroken = false;
}
*/

void CVXS_Voxel::CalcNearby(int NumHops) //populates Nearby
{
	NearbyVoxInds.clear();
	int StartPoint = 0; //our enter and exit point (so we don't repeat for each iteration
	int StopPoint = 1;

	NearbyVoxInds.push_back(MySIndex);

	for (int i=0; i<NumHops; i++){
		for (int j=StartPoint; j<StopPoint; j++){ //go through the list from the most recent interation...
			for (int k=0; k<pNearbyVox(j)->NumLocalBonds(); k++){ //look at all the bonds of this voxel
				//get the other voxel in this bond...
				int OtherSIndex = pNearbyVox(j)->pBond(k)->Vox1()->MySIndex;
				if (pNearbyVox(j)->IsMe(pNearbyVox(j)->pBond(k)->Vox1()))  OtherSIndex = pNearbyVox(j)->pBond(k)->Vox2()->MySIndex; //if this voxel 1

				//Add it to the list if its not already on it.
				if (!IsNearbyVox(OtherSIndex)) NearbyVoxInds.push_back(OtherSIndex);
			}
		}
		StartPoint = StopPoint;
		StopPoint = NumNearbyVox();
	}
}

void CVXS_Voxel::CalcVoxMatStress(vfloat StrainIn, vfloat* pStressOut, bool* IsPastYielded, bool* IsPastFail)
{
	if (!p_Sim->BlendingEnabled){ //if we're not doing any material blending
		p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetModelStress(StrainIn, pStressOut, IsPastYielded, IsPastFail);
	}
	else {
		vfloat Sum = 0;
		vfloat tmpWeight = 0;
		vfloat tmpStress;
		bool tmpYield, tmpFail;
		*IsPastYielded = false;
		*IsPastFail = false;


		//ONLY WORKS WITH TWO MATERIALS OAT A TIME! MAJOR LIMITATION IN NEED OF WORK!
		vfloat MinStress= FLT_MAX;
		vfloat MaxStress = -FLT_MAX;
		vfloat PercStronger = 0;
		//get min and max
		for (int MatIndex = 1; MatIndex<p_Sim->LocalVXC.GetNumMaterials(); MatIndex++){
			if (BlendMix[MatIndex] != 0){
				p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetModelStress(StrainIn, &tmpStress, &tmpYield, &tmpFail);
				if(tmpYield) *IsPastYielded = true; //yielded if any material yielded
				if(tmpFail) *IsPastFail = true; //yielded if any material yielded


				if (tmpStress<MinStress){
					MinStress=tmpStress;
				}
				if (tmpStress>MaxStress){
					PercStronger=BlendMix[MatIndex];
					MaxStress=tmpStress;
				}
			}
		}
		if (MinStress == MaxStress){*pStressOut=MinStress; *IsPastYielded=tmpYield; *IsPastFail=tmpFail; return;}

		switch (p_Sim->BlendModel){
			case MB_LINEAR: tmpWeight = PercStronger; break;
			case MB_EXPONENTIAL: tmpWeight = pow(2, PercStronger)-1; break;
			case MB_POLYNOMIAL: tmpWeight = pow(PercStronger, p_Sim->PolyExp); break;
			}

		*pStressOut = tmpWeight*(MaxStress-MinStress)+MinStress;


		/*
		for (int MatIndex = 1; MatIndex<p_Sim->LocalVXC.GetNumMaterials(); MatIndex++){
			if (BlendMix[MatIndex] != 0){ //if contains any of this material
				p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetModelStress(StrainIn, &tmpStress, &tmpYield, &tmpFail);
				if(tmpYield) *IsPastYielded = true; //yielded if any material yielded
				if(tmpFail) *IsPastFail = true; //yielded if any material yielded

				vfloat Perc = (tmpStress-MinStress)/(MaxStress-MinStress);

				switch (p_Sim->BlendModel){
				case MB_LINEAR: tmpWeight = Perc; break;
				case MB_EXPONENTIAL: tmpWeight = pow(2, Perc)-1; break;
				case MB_POLYNOMIAL: tmpWeight = pow(Perc, p_Sim->PolyExp); break;
				}
				Sum += BlendMix[MatIndex]*(tmpWeight*(MaxStress-MinStress)+MinStress);
			}
		}
		*pStressOut = Sum;
		*/
	}

}

vfloat CVXS_Voxel::GetEMod(void) //gets elastic modulus, taking blending into account
{
	if (!p_Sim->BlendingEnabled){ //if we're not doing any material blending
		return p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetElasticMod();
	}

	//cache this!!!

	else { //return combo of elastic moduli according to blending model
		vfloat Sum = 0;
		vfloat tmpWeight = 0;
		vfloat PercWeaker = 0;

		//ONLY WORKS WITH TWO MATERIALS OAT A TIME! MAJOR LIMITATION IN NEED OF WORK!
		vfloat MinEMod = FLT_MAX;
		vfloat MaxEMod = 0;
		//get min and max
		for (int MatIndex = 1; MatIndex<p_Sim->LocalVXC.GetNumMaterials(); MatIndex++){
			if (BlendMix[MatIndex] != 0){
				vfloat ThisEMod = p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetElasticMod();
				if (ThisEMod<MinEMod){
					PercWeaker = BlendMix[MatIndex];
					MinEMod=ThisEMod;
				}
				if (ThisEMod>MaxEMod) MaxEMod=ThisEMod;
			}
		}
		if (MinEMod == MaxEMod) return MaxEMod;


		switch (p_Sim->BlendModel){
			case MB_LINEAR: tmpWeight = PercWeaker; break;
			case MB_EXPONENTIAL: tmpWeight = pow(2, PercWeaker)-1; break;
			case MB_POLYNOMIAL: tmpWeight = pow(PercWeaker, p_Sim->PolyExp); break;
			}

		return tmpWeight*(MaxEMod-MinEMod)+MinEMod;

		/*
		for (int MatIndex = 1; MatIndex<p_Sim->LocalVXC.GetNumMaterials(); MatIndex++){
			if (BlendMix[MatIndex] != 0){ //if contains any of this material
				vfloat ThisEMod = p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetElasticMod();
				vfloat Perc = (ThisEMod-MinEMod)/(MaxEMod-MinEMod);

				switch (p_Sim->BlendModel){
				case MB_LINEAR: tmpWeight = Perc; break;
				case MB_EXPONENTIAL: tmpWeight = pow(2, Perc)-1; break;
				case MB_POLYNOMIAL: tmpWeight = pow(Perc, p_Sim->PolyExp); break;
				}
				Sum += BlendMix[MatIndex]*(tmpWeight*(MaxEMod-MinEMod)+MinEMod);
			}
		}
		return Sum;
		*/
	}
}

vfloat CVXS_Voxel::GetPoisson(void) //gets poisson ratio, taking blending into account
{
	if (!p_Sim->BlendingEnabled){ //if we're not doing any material blending
		return p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetPoissonsRatio();
	}

	//cache this!!!
	else { //for now, return linear average Poissons ratio
		vfloat Sum = 0;
		for (int MatIndex = 1; MatIndex<p_Sim->LocalVXC.GetNumMaterials(); MatIndex++){
			if (BlendMix[MatIndex] != 0){ //if contains any of this material
				Sum += BlendMix[MatIndex]*p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetPoissonsRatio();
			}
		}
		return Sum; //Don't have to normalize by weight since BlendMix is already normalized.
	}
}

vfloat CVXS_Voxel::GetCTE(void) //gets CTE, taking blending into account
{
	if (!p_Sim->BlendingEnabled){ //if we're not doing any material blending
		return p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetCTE();
	}


	//cache this!!!
	else { //for now, return linear average CTE
		vfloat Sum = 0;
		for (int MatIndex = 1; MatIndex<p_Sim->LocalVXC.GetNumMaterials(); MatIndex++){
			if (BlendMix[MatIndex] != 0){ //if contains any of this material
				Sum += BlendMix[MatIndex]*p_Sim->LocalVXC.GetBaseMat(MatIndex)->GetCTE();
			}
		}
		return Sum; //Don't have to normalize by weight since BlendMix is already normalized.
	}
}

void CVXS_Voxel::CalcMyBlendMix() //requires MyXIndex, p_sim and p_Sim->LocalVXC to already be in place...
{
	//initialize blendmix to size of palette with zeros...
	BlendMix.clear();
	BlendMix.resize(p_Sim->LocalVXC.GetNumMaterials(), 0);

	vfloat BlendRad=p_Sim->MixRadius; //make a local copy for convenience
	int x, y, z, ThisInd;
	Vec3D ThisPos;
	vfloat ThisDist, ThisWeight;
	p_Sim->LocalVXC.GetXYZNom(&x, &y, &z, MyXIndex);
	Vec3D BasePos = p_Sim->LocalVXC.GetXYZ(MyXIndex);
	Vec3D EnvMult = p_Sim->LocalVXC.Lattice.GetDimAdj();
	int xLook = 3*BlendRad/EnvMult.x+1;
	int yLook = 3*BlendRad/EnvMult.y+1;
	int zLook = 3*BlendRad/EnvMult.z+1;

	//BlendRad is the standard deviation of the gaussian mixing weighting . (multiply by GetLatticeDim to get the real-valued standard dev)
//	vfloat AvgEMod = 0.0;
	vfloat TotalWeight = 0.0;

	//up to three sigma...
	for (int ix = x-xLook; ix<=x+xLook; ix++){
		for (int jy = y-yLook; jy<=y+yLook; jy++){
			for (int kz = z-zLook; kz<=z+zLook; kz++){
				ThisInd = p_Sim->LocalVXC.GetIndex(ix, jy, kz);
				if (ThisInd == -1) continue; //if invalid location then skip
				if (p_Sim->LocalVXC.GetMat(ThisInd) == 0) continue; //skip this one if it is empty (no material...
					
				ThisPos = p_Sim->LocalVXC.GetXYZ(ThisInd);
				ThisDist = ((ThisPos-BasePos).Length())/p_Sim->LocalVXC.GetLatticeDim();

				ThisWeight = exp(-0.5*ThisDist*ThisDist/(BlendRad*BlendRad));
				BlendMix[p_Sim->LocalVXC.GetLeafMatIndex(ThisInd)] += ThisWeight;
				TotalWeight += ThisWeight;
			}
		}
	}

	//normalize to make sure BlendMix adds up to 1!
	for(std::vector<vfloat>::iterator it = BlendMix.begin(); it != BlendMix.end(); it++)
		*it /= TotalWeight;

}
