/* ----------------------------------------------------------------------------
  ex21.C
  mbwall 1jan96
  Copyright (c) 1995-1996 Massachusetts Institute of Technology

 DESCRIPTION:
   This example illustrates various uses of the AlleleSet in concert with the
ArrayAlleleGenome.  In particular, we use the RealGenome to show how you can
use the enumerated and bounded types of allele sets.
   You can define one allele set for an entire array, or you can define one
allele set for each element in the array.  The constructor that you use to 
create the array determines which behaviour you'll get.
---------------------------------------------------------------------------- */
#include <stdio.h>
#include <ga/ga.h>
#include <ga/std_stream.h>
#define TIXML_USE_STL
//#include <tinyxml.h> // #include “tinyxml.h"?

#define cout STD_COUT
#define endl STD_ENDL

#define INSTANTIATE_REAL_GENOME
#include <ga/GARealGenome.h>

#include <ga/GA2DArrayGenome.h>
#include <ga/GARealGenome.h>
#include <ga/Array3D.h>  //#include <hiller/std::vector3D.h
#include <iostream>
#include <sstream>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <string>
#include <queue>
#include <list>
#include <vector>
#include <math.h>
using namespace std;

template<class InputIterator, class T>
  InputIterator find ( InputIterator first, InputIterator last, const T& value )
  {
    for ( ;first!=last; first++) if ( *first==value ) break;
    return first;
  }

int num_x_voxels;
int num_y_voxels;
int num_z_voxels;

int length;

int indNum;

float VoxelSize;

float Objective2(GAGenome &);
//float WriteChamp(GAGenome);
int writeVoxelyzeFile( std::vector<float> ContinuousArray, std::vector<float> PassiveStiffArray, std::vector<float> PassiveSoftArray,std::vector<float> Phase1Array , std::vector<float> Phase2Array, string individualID);



int
main(int argc, char** argv)
{
  cout << "Example 21\n\n";
  cout << "This example shows various uses of the allele set object\n";
  cout << "in combination with the real number genome.\n\n"; cout.flush();

// See if we've been given a seed to use (for testing purposes).  When you
// specify a random seed, the evolution will be exactly the same each time
// you use that seed number.

  unsigned int seed = 0;
  for(int ii=1; ii<argc; ii++) {
    if(strcmp(argv[ii++],"seed") == 0) {
      seed = atoi(argv[ii]);
    }
  }

// First make a bunch of genomes.  We'll use each one in turn for a genetic
// algorithm later on.  Each one illustrates a different method of using the
// allele set object.  Each has its own objective function.


num_x_voxels = 10;
num_y_voxels = 10;
num_z_voxels = 10;

indNum=0;

length = num_x_voxels*num_x_voxels*num_x_voxels;

VoxelSize = /*FixedBoundingBoxSize*/10 * 0.001/num_x_voxels;

// This genome uses a bounded set of continous numbers.  The default arguments
// are INCLUSIVE for both the lower and upper bounds, so in this case the 
// allele set is [0,1] and any element of the genome may assume a value [0,1].

  GARealAlleleSet alleles2(0, 1);
  GARealGenome genome2(length*5, alleles2, Objective2);


// Now that we have the genomes, create a parameter list that will be used for
// all of the genetic algorithms and all of the genomes.

  GAParameterList params;
  GASteadyStateGA::registerDefaultParameters(params);
  params.set(gaNnGenerations, 1000);
  params.set(gaNpopulationSize, 30);
  params.set(gaNscoreFrequency, 1);
  params.set(gaNpMutation, 0.001);
  params.set(gaNpCrossover, 0);
  params.set(gaNselectScores, (int)GAStatistics::AllScores);
  params.parse(argc, argv, gaFalse);


// Now do a genetic algorithm for each one of the genomes that we created.


  GASteadyStateGA ga2(genome2);
  ga2.parameters(params);
  ga2.set(gaNscoreFilename, "stats.dat");
  cout << "\nStarting Direct Encoding Softobts Stuffs..." << endl << endl;
  ga2.evolve();
  cout << "the ga generated:\n" << ga2.statistics().bestIndividual() << endl;

  cout << endl << "GA stats: " << ga2.statistics() << endl;

  //WriteChamp(ga2.statistics().bestIndividual());
	std::ofstream myfile;
  	myfile.open ("winner.txt");
	myfile << ga2.statistics().bestIndividual();
	myfile.close();

	std::ifstream fitfile;
	fitfile.open("winner.txt");
	std::string line;
	string theline;
	int j=0;

	std::vector<float> ContinuousArray (length,0.0);
	std::vector<float> PassiveStiffArray (length,0.0);
	std::vector<float> PassiveSoftArray (length,0.0);
	std::vector<float> Phase1Array (length,0.0);
	std::vector<float> Phase2Array (length,0.0);
	
	while (std::getline(fitfile, line, ' '))
	{
		//cout << j << endl;
		//cout << line << endl;
		//if (j==1) { theline = line; }

		j++;
		if (j<length) { ContinuousArray[j] = atof(line.c_str()); }
		if (j>=length && j<2*length) { PassiveStiffArray[j-length] = atof(line.c_str()); } 
		if (j>=2*length && j<3*length) { PassiveSoftArray [j-2*length] = atof(line.c_str()); } 
		if (j>=3*length && j<4*length) { Phase1Array [j-3*length] = atof(line.c_str()); } 
		if (j>=4*length && j<5*length) { Phase2Array [j-4*length] = atof(line.c_str()); } 
	}

	string individualID = "champ";
	writeVoxelyzeFile(ContinuousArray, PassiveStiffArray, PassiveSoftArray, Phase1Array , Phase2Array, individualID);

  return 0;
}
 


// This objective function tries to generate a straight - it gives higher score
// to a genome whose elements descend in value.  If two genomes both have
// elements in strictly descending order, they get the same score regardless
// of their values.

float
WriteChamp(GAGenome g)
{
  GARealGenome& genome = (GARealGenome&)g;

//cout << genome << endl;
float fitness = 0.0001;

std::vector<float> ContinuousArray (length,0.0);
std::vector<float> PassiveStiffArray (length,0.0);
std::vector<float> PassiveSoftArray (length,0.0);
std::vector<float> Phase1Array (length,0.0);
std::vector<float> Phase2Array (length,0.0);

  float value=0.0;
  for(int i=0; i<length; i++) {
    //if(genome.gene(i) < genome.gene(i-1)) value += 1.0;
	ContinuousArray[i] = genome.gene(i);
	PassiveStiffArray[i] = genome.gene(i+length);
	PassiveSoftArray[i] = genome.gene(i+2*length);
	Phase1Array[i] = genome.gene(i+3*length);
	Phase2Array[i] = genome.gene(i+4*length);
  }
/*
		std::ostringstream num;
		num << indNum;
		string individualID = num.str();
*/

		//indNum++;

string individualID = "champ";

 	writeVoxelyzeFile(ContinuousArray, PassiveStiffArray, PassiveSoftArray, Phase1Array , Phase2Array, individualID);

}


float
Objective2(GAGenome& g)
{
  GARealGenome& genome = (GARealGenome&)g;
//  GASteadyStateGA ga2 = genome.geneticAlgorithm();

//cout << genome << endl;
float fitness = 0.0001;

std::vector<float> ContinuousArray (length,0.0);
std::vector<float> PassiveStiffArray (length,0.0);
std::vector<float> PassiveSoftArray (length,0.0);
std::vector<float> Phase1Array (length,0.0);
std::vector<float> Phase2Array (length,0.0);

  float value=0.0;
  for(int i=0; i<length; i++) {
    //if(genome.gene(i) < genome.gene(i-1)) value += 1.0;
	ContinuousArray[i] = genome.gene(i);
	PassiveStiffArray[i] = genome.gene(i+length);
	PassiveSoftArray[i] = genome.gene(i+2*length);
	Phase1Array[i] = genome.gene(i+3*length);
	Phase2Array[i] = genome.gene(i+4*length);
  }
/*
		std::ostringstream num;
		num << indNum;
		string individualID = num.str();
*/

		indNum++;

string individualID = "test";

 	writeVoxelyzeFile(ContinuousArray, PassiveStiffArray, PassiveSoftArray, Phase1Array , Phase2Array, individualID);

	cout << "Staring Eval for individual " << indNum << endl;

	std::ostringstream callVoxleyzeCmd;
	callVoxleyzeCmd << "time ~/code/softbots/Voxel_Sim_NoGui/voxelize -f " << "test" << "_genome.vxa";
		
	int exitCode = std::system(callVoxleyzeCmd.str().c_str());


/* read xml fitness file */
/*
				TiXmlDocument doc;
				std::ostringstream outFileName;
				outFileName << individualID << "_fitness.xml";   		// <-- for real
				//outFileName << "nan_fitness.xml";						// <-- debug only

				if(!doc.LoadFile(outFileName.str().c_str()))
				{
					cerr << doc.ErrorDesc() << endl;
					return 0.000001; //return FAILURE;
				}	
		
				TiXmlElement* root = doc.FirstChildElement();
				if(root == NULL)
				{
					cerr << "Failed to load file: No root element."
						 << endl;
					doc.Clear();
					return 0.001; //return FAILURE;
				}

				fitness = 0.001;

				for(TiXmlElement* elem1 = root->FirstChildElement(); elem1 != NULL; elem1 = elem1->NextSiblingElement())
				{
			
					for(TiXmlElement* elem = elem1->FirstChildElement(); elem != NULL; elem = elem->NextSiblingElement())
					{

						string elemName = elem->Value();

			
						for(TiXmlNode* e = elem->FirstChild(); e != NULL; e = e->NextSibling())
						{

							cout << "got here 1" << endl;
							TiXmlText* text = e->ToText();
							cout << "element from voxelyze: " << e->Value() << endl;
							cout << "raw fitness value from voxelyze: " << fabs(atof(e->Value())) << endl;
							cout << "body length fitness value from voxelyze: " << fabs(atof(e->Value()))/(num_x_voxels*VoxelSize) << endl;
							fitness = fabs(atof(e->Value()))/(num_x_voxels*VoxelSize);

							if (std::string::npos != string(e->Value()).find('n'))
							{
							  fitness = 0.0001;
							  cout << "THIS INDIVIDUAL RETURNED NaN... setting arbitrary low fitness" << endl;
							}

						}
					}
				}
				doc.Clear();
*/

	std::ifstream fitfile;
	fitfile.open("test_fitness.xml");
	std::string line;
	string theline;
	int j=0;
	while (std::getline(fitfile, line))
	{
		//cout << j << endl;
		//cout << line << endl;
		if (j==3) { theline = line; }

		j++;
	}
	
	theline = theline.substr(23,theline.rfind("<")-23);

	if 	(std::string::npos != string(theline).find('n')) 
	{ 
		fitness = 0.0001; 
	} else 
	{
		fitness = fabs(atof(theline.c_str()))/(num_x_voxels*VoxelSize);
		if (fitness < 0.0001 || fitness > 1000) { fitness = 0.0001; } 
	} 
	 

	cout << "string from voxelyze: " << theline << endl;
	cout << "fitness: " << fitness << endl << endl;

//	cout << "best so far at generation "<< ga2.statistics().generation() << ": " << ga2.statistics.maxEver()  << endl << endl;


  return fitness;
}




// from SoftbotsExperiment.cpp:


	int leftExists(std::vector<int> ArrayForVoxelyze, int currentPosition)
	{
		
//cout << "ArrayForVoxelyze.GetFullSize(): "<< ArrayForVoxelyze.GetFullSize() << endl;
				if (currentPosition % num_x_voxels == 0) 
				{
					return -999;	
				}
				else if (ArrayForVoxelyze[currentPosition - 1] == 0)
				{
					 return -999;
				}

				else
				{
				return (currentPosition - 1);
				}
	}

	int rightExists(std::vector<int> ArrayForVoxelyze, int currentPosition)
	{

				if (currentPosition % num_x_voxels == num_x_voxels-1) return -999;				
				else if (ArrayForVoxelyze[currentPosition + 1] == 0) return -999;
				else {/*numConnections++;*/ return (currentPosition + 1);}
	}
			
	int forwardExists(std::vector<int> ArrayForVoxelyze, int currentPosition)
	{


				if (int(currentPosition/num_x_voxels) % num_y_voxels == num_y_voxels-1) return -999;				
				else if (ArrayForVoxelyze[currentPosition + num_x_voxels] == 0) return -999;
				else {/*numConnections++;*/ return (currentPosition + num_x_voxels);}
	}

	int backExists(std::vector<int> ArrayForVoxelyze, int currentPosition)
	{


				if (int(currentPosition/num_x_voxels) % num_y_voxels == 0) return -999;				
				else if (ArrayForVoxelyze[currentPosition - num_x_voxels] == 0) return -999;
				else {/*numConnections++;*/ return (currentPosition - num_x_voxels);}
	}

	int upExists(std::vector<int> ArrayForVoxelyze, int currentPosition)
	{
	

				if (int(currentPosition/(num_x_voxels*num_y_voxels)) % num_z_voxels == num_z_voxels-1) return -999;				
				else if (ArrayForVoxelyze[currentPosition + num_x_voxels*num_y_voxels] == 0) return -999;
				else {/*numConnections++;*/ return (currentPosition + num_x_voxels*num_y_voxels);}
	}

	int downExists(std::vector<int> ArrayForVoxelyze, int currentPosition)
	{


				if (int(currentPosition/(num_x_voxels*num_y_voxels)) % num_z_voxels == 0) return -999;				
				else if (ArrayForVoxelyze[currentPosition - num_x_voxels*num_y_voxels] == 0) return -999;
				else {/*numConnections++;*/ return (currentPosition - num_x_voxels*num_y_voxels);}
	}


	std::pair< std::queue<int>, std::list<int> > circleOnce(std::vector<int> ArrayForVoxelyze, std::queue<int> queueToCheck, std::list<int> alreadyChecked)
	{

				int currentPosition = queueToCheck.front();

				queueToCheck.pop();

				int index;

				index = leftExists(ArrayForVoxelyze, currentPosition);

				if (index>0 && find(alreadyChecked.begin(),alreadyChecked.end(),index)==alreadyChecked.end())
				{queueToCheck.push(index);}
				alreadyChecked.push_back(index);
				index = rightExists(ArrayForVoxelyze, currentPosition);

				if (index>0 && find(alreadyChecked.begin(),alreadyChecked.end(),index)==alreadyChecked.end())
				{queueToCheck.push(index);}
				alreadyChecked.push_back(index);
				index = forwardExists(ArrayForVoxelyze, currentPosition);

				if (index>0 && find(alreadyChecked.begin(),alreadyChecked.end(),index)==alreadyChecked.end())
				{queueToCheck.push(index);}
				alreadyChecked.push_back(index);
				index = backExists(ArrayForVoxelyze, currentPosition);

				if (index>0 && find(alreadyChecked.begin(),alreadyChecked.end(),index)==alreadyChecked.end())
				{queueToCheck.push(index);}
				alreadyChecked.push_back(index);
				index = upExists(ArrayForVoxelyze, currentPosition);

				if (index>0 && find(alreadyChecked.begin(),alreadyChecked.end(),index)==alreadyChecked.end())
				{queueToCheck.push(index);}
				alreadyChecked.push_back(index);
				index = downExists(ArrayForVoxelyze, currentPosition);

				if (index>0 && find(alreadyChecked.begin(),alreadyChecked.end(),index)==alreadyChecked.end())
				{queueToCheck.push(index);}
				alreadyChecked.push_back(index);
				return make_pair (queueToCheck, alreadyChecked);
	}

	std::vector<int> makeOneShapeOnly(std::vector<int> ArrayForVoxelyze)
	{
			//find start value:
			std::queue<int> queueToCheck;
			std::list<int> alreadyChecked;
			std::pair<  std::queue<int>, std::list<int> > queueAndList;
			int startVoxel = int(ArrayForVoxelyze.size()/2);
			queueToCheck.push(startVoxel);
			alreadyChecked.push_back(startVoxel);

			while (ArrayForVoxelyze[startVoxel] == 0 && not queueToCheck.empty())
				{
			
					startVoxel = queueToCheck.front();
					queueAndList = circleOnce(ArrayForVoxelyze, queueToCheck, alreadyChecked);
					queueToCheck = queueAndList.first;
					alreadyChecked = queueAndList.second;
					
				}

			// create only one shape:	
			alreadyChecked.clear();
			alreadyChecked.push_back(startVoxel);
			//queueToCheck.clear();
			while (!queueToCheck.empty()) {queueToCheck.pop();}
			queueToCheck.push(startVoxel);
			while (!queueToCheck.empty())		
			{
				startVoxel = queueToCheck.front();

				queueAndList = circleOnce(ArrayForVoxelyze, queueToCheck, alreadyChecked);

				queueToCheck = queueAndList.first;
				alreadyChecked = queueAndList.second;

			}

			for (int v=0; v<ArrayForVoxelyze.size(); v++)
			{
				if (find(alreadyChecked.begin(),alreadyChecked.end(),v)==alreadyChecked.end())
				{
					ArrayForVoxelyze[v]=0;
				}
			}
		return ArrayForVoxelyze;
	}


	std::vector<int> createArrayForVoxelyze(std::vector<float> ContinuousArray ,std::vector<float> PassiveStiffArray, std::vector<float> PassiveSoftArray ,std::vector<float> Phase1Array, std::vector<float> Phase2Array)
	{	


		std::vector<int> ArrayForVoxelyze (length,0.0); //std::vector denoting type of material

		int numVoxelsFilled   = 0;
		int numVoxelsActuated = 0;
		int numConnections    = 0;
		int v=0;

		// nac: hard coded numMaterials
		int numMaterials = 4;

		for (int z=0; z<num_z_voxels; z++)
		{
			for (int y=0; y<num_y_voxels; y++)
			{
				for (int x=0; x<num_x_voxels; x++)
				{
					// NOTE: this uses mateiral 0 as passive, material 1 as active phase 1, and material 2 as active phase 2
		
					if(ContinuousArray[v] < 0.5)
					{ ArrayForVoxelyze[v] = 0;}
					else if(numMaterials > 1 && PassiveSoftArray[v]>=Phase2Array[v] && PassiveSoftArray[v]>=Phase1Array[v] && PassiveSoftArray[v]>=PassiveStiffArray[v]) 
					{ ArrayForVoxelyze[v] = 2;}
					else if(numMaterials > 3 && PassiveStiffArray[v]>=Phase2Array[v] && PassiveStiffArray[v]>=Phase1Array[v] && PassiveStiffArray[v]>=PassiveSoftArray[v]) 
					{ ArrayForVoxelyze[v] = 4;}
					else if(numMaterials > 0 && Phase1Array[v]>=Phase2Array[v] && Phase1Array[v]>=PassiveSoftArray[v] && Phase1Array[v]>=PassiveStiffArray[v]) 
					{ ArrayForVoxelyze[v] = 1;}
					else if(numMaterials > 2 && Phase2Array[v]>=Phase1Array[v] && Phase2Array[v]>=PassiveSoftArray[v] && Phase2Array[v]>=PassiveStiffArray[v]) 
					{ ArrayForVoxelyze[v] = 3;}
					else 
					{ 
						if (numMaterials > 1) {ArrayForVoxelyze[v] = 2;}
						else {ArrayForVoxelyze[v] = 1;}
					} //if a tie, just use soft passive
		 
					// change "=" to random assignment? currently biases passive, then phase 1, then phase 2 if all equal

					v++;
				}
			}
		}

		ArrayForVoxelyze = makeOneShapeOnly(ArrayForVoxelyze);

		return ArrayForVoxelyze;
	}




// from SoftbotsExperiment.cpp:

    int writeVoxelyzeFile( std::vector<float> ContinuousArray, std::vector<float> PassiveStiffArray, std::vector<float> PassiveSoftArray,std::vector<float> Phase1Array , std::vector<float> Phase2Array, string individualID)
	{

		std::ofstream md5file;
		md5file.open ("md5sumTMP.txt");		

		std::ostringstream stiffness;
		stiffness << 50;//(int(NEAT::Globals::getSingleton()->getParameterValue("MaxStiffness")));


  		std::ofstream myfile;
		std::ostringstream myFileName;
		myFileName << individualID << "_genome.vxa";
  		myfile.open (myFileName.str().c_str());
//1000000000
  		myfile << "\
<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n\
<VXA Version=\"1.0\">\n\
<Simulator>\n\
<Integration>\n\
<Integrator>0</Integrator>\n\
<DtFrac>0.9</DtFrac>\n\
</Integration>\n\
<Damping>\n\
<BondDampingZ>1</BondDampingZ>\n\
<ColDampingZ>0.8</ColDampingZ>\n\
<SlowDampingZ>0.001</SlowDampingZ>\n\
</Damping>\n\
<Collisions>\n\
<SelfColEnabled>1</SelfColEnabled>\n\
<ColSystem>3</ColSystem>\n\
<CollisionHorizon>2</CollisionHorizon>\n\
</Collisions>\n\
<Features>\n\
<FluidDampEnabled>0</FluidDampEnabled>\n\
<PoissonKickBackEnabled>0</PoissonKickBackEnabled>\n\
<EnforceLatticeEnabled>0</EnforceLatticeEnabled>\n\
</Features>\n\
<SurfMesh>\n\
<CMesh>\n\
<DrawSmooth>1</DrawSmooth>\n\
<Vertices/>\n\
<Facets/>\n\
<Lines/>\n\
</CMesh>\n\
</SurfMesh>\n\
<SimStop>\n\
<SimStopTime>0.25</SimStopTime>\n\
<SimStopStep>1000000000</SimStopStep>\n\
</SimStop>\n\
<Fitness>\n\
<!--Track the center of mass -->\n\
<TrackCOM>0</TrackCOM>\n\
<!-- Specify the fitness filename that will be output -->\n\
<FitnessFileNm>" << individualID << "_fitness.xml</FitnessFileNm>\n\
</Fitness>\n\
</Simulator>\n\
<Environment>\n\
<Fixed_Regions>\n\
<NumFixed>0</NumFixed>\n\
</Fixed_Regions>\n\
<Forced_Regions>\n\
<NumForced>0</NumForced>\n\
</Forced_Regions>\n\
<Gravity>\n\
<GravEnabled>1</GravEnabled>\n\
<GravAcc>-27.468</GravAcc>\n\
<FloorEnabled>1</FloorEnabled>\n\
</Gravity>\n\
<Thermal>\n\
<TempEnabled>1</TempEnabled>\n\
<TempAmp>39</TempAmp>\n\
<TempBase>25</TempBase>\n\
<VaryTempEnabled>1</VaryTempEnabled>\n\
<TempPeriod>0.025</TempPeriod>\n\
</Thermal>\n\
</Environment>\n\
<VXC Version=\"0.93\">\n\
<Lattice>\n\
<Lattice_Dim>" << VoxelSize << "</Lattice_Dim>\n\
<X_Dim_Adj>1</X_Dim_Adj>\n\
<Y_Dim_Adj>1</Y_Dim_Adj>\n\
<Z_Dim_Adj>1</Z_Dim_Adj>\n\
<X_Line_Offset>0</X_Line_Offset>\n\
<Y_Line_Offset>0</Y_Line_Offset>\n\
<X_Layer_Offset>0</X_Layer_Offset>\n\
<Y_Layer_Offset>0</Y_Layer_Offset>\n\
</Lattice>\n\
<Voxel>\n\
<Vox_Name>BOX</Vox_Name>\n\
<X_Squeeze>1</X_Squeeze>\n\
<Y_Squeeze>1</Y_Squeeze>\n\
<Z_Squeeze>1</Z_Squeeze>\n\
</Voxel>\n\
<Palette>\n\
<Material ID=\"1\">\n\
<MatType>0</MatType>\n\
<Name>Active_+</Name>\n\
<Display>\n\
<Red>1</Red>\n\
<Green>0</Green>\n\
<Blue>0</Blue>\n\
<Alpha>1</Alpha>\n\
</Display>\n\
<Mechanical>\n\
<MatModel>0</MatModel>\n\
<Elastic_Mod>1e+007</Elastic_Mod>\n\
<Plastic_Mod>0</Plastic_Mod>\n\
<Yield_Stress>0</Yield_Stress>\n\
<FailModel>0</FailModel>\n\
<Fail_Stress>0</Fail_Stress>\n\
<Fail_Strain>0</Fail_Strain>\n\
<Density>1e+006</Density>\n\
<Poissons_Ratio>0.35</Poissons_Ratio>\n\
<CTE>0.01</CTE>\n\
<uStatic>1</uStatic>\n\
<uDynamic>0.5</uDynamic>\n\
</Mechanical>\n\
</Material>\n\
<Material ID=\"2\">\n\
<MatType>0</MatType>\n\
<Name>Passive_Soft</Name>\n\
<Display>\n\
<Red>0</Red>\n\
<Green>1</Green>\n\
<Blue>1</Blue>\n\
<Alpha>1</Alpha>\n\
</Display>\n\
<Mechanical>\n\
<MatModel>0</MatModel>\n\
<Elastic_Mod>1e+007</Elastic_Mod>\n\
<Plastic_Mod>0</Plastic_Mod>\n\
<Yield_Stress>0</Yield_Stress>\n\
<FailModel>0</FailModel>\n\
<Fail_Stress>0</Fail_Stress>\n\
<Fail_Strain>0</Fail_Strain>\n\
<Density>1e+006</Density>\n\
<Poissons_Ratio>0.35</Poissons_Ratio>\n\
<CTE>0</CTE>\n\
<uStatic>1</uStatic>\n\
<uDynamic>0.5</uDynamic>\n\
</Mechanical>\n\
</Material>\n\
<Material ID=\"3\">\n\
<MatType>0</MatType>\n\
<Name>Active_-</Name>\n\
<Display>\n\
<Red>0</Red>\n\
<Green>1</Green>\n\
<Blue>0</Blue>\n\
<Alpha>1</Alpha>\n\
</Display>\n\
<Mechanical>\n\
<MatModel>0</MatModel>\n\
<Elastic_Mod>1e+007</Elastic_Mod>\n\
<Plastic_Mod>0</Plastic_Mod>\n\
<Yield_Stress>0</Yield_Stress>\n\
<FailModel>0</FailModel>\n\
<Fail_Stress>0</Fail_Stress>\n\
<Fail_Strain>0</Fail_Strain>\n\
<Density>1e+006</Density>\n\
<Poissons_Ratio>0.35</Poissons_Ratio>\n\
<CTE>-0.01</CTE>\n\
<uStatic>1</uStatic>\n\
<uDynamic>0.5</uDynamic>\n\
</Mechanical>\n\
</Material>\n\
<Material ID=\"4\">\n\
<MatType>0</MatType>\n\
<Name>Passive_Hard</Name>\n\
<Display>\n\
<Red>0</Red>\n\
<Green>0</Green>\n\
<Blue>1</Blue>\n\
<Alpha>1</Alpha>\n\
</Display>\n\
<Mechanical>\n\
<MatModel>0</MatModel>\n\
<Elastic_Mod>50e+006</Elastic_Mod>\n\
<Plastic_Mod>0</Plastic_Mod>\n\
<Yield_Stress>0</Yield_Stress>\n\
<FailModel>0</FailModel>\n\
<Fail_Stress>0</Fail_Stress>\n\
<Fail_Strain>0</Fail_Strain>\n\
<Density>1e+006</Density>\n\
<Poissons_Ratio>0.35</Poissons_Ratio>\n\
<CTE>0</CTE>\n\
<uStatic>1</uStatic>\n\
<uDynamic>0.5</uDynamic>\n\
</Mechanical>\n\
</Material>\n\
</Palette>\n\
<Structure Compression=\"ASCII_READABLE\">\n\
<X_Voxels>" << num_x_voxels << "</X_Voxels>\n\
<Y_Voxels>" << num_y_voxels << "</Y_Voxels>\n\
<Z_Voxels>" << num_z_voxels << "</Z_Voxels>\n\
<Data>\n\
<Layer><![CDATA[";

std::vector<int> ArrayForVoxelyze = createArrayForVoxelyze(ContinuousArray, PassiveStiffArray, PassiveSoftArray, Phase1Array, Phase2Array);

/* nac: marching cubes, finish this! /

std::cout << "Performing marching cubes...\n";
CMesh OutputMesh;
CMarchCube::SingleMaterialMultiColor(&OutputMesh, &ContinuousArray, &rColorstd::vector, &gColorstd::vector, &bColorstd::vector, voxelExistsThreshold, VoxelSize*1000);	//jmc: last argument is the threshold above which we consider the voxel to be extant

//format: 	static void MultiMaterial(CMesh* pMeshOut, void* pstd::vectors, bool SumMat, CColor* pColors = NULL, float Thresh = 0.0f, float Scale = 1.0f);
*/

int v=0;
for (int z=0; z<num_z_voxels; z++)
{
	for (int y=0; y<num_y_voxels; y++)
	{
		for (int x=0; x<num_x_voxels; x++)
		{
			// NOTE: this uses mateiral 0 as passive, material 1 as active phase 1, and material 2 as active phase 2
/*
			if(ContinuousArray[v] >= Phase1Array[v] && ContinuousArray[v] >= Phase2Array[v] && ContinuousArray[v] >= Passivestd::vector[v]) myfile << 0;
			else if(Passivestd::vector[v] >= ContinuousArray[v] && Passivestd::vector[v] >= Phase2Array[v] && Passivestd::vector[v] >= Phase1Array[v]) myfile << 1;
			else if(Phase1Array[v] >= ContinuousArray[v] && Phase1Array[v] >= Phase2Array[v] && Phase1Array[v] >= Passivestd::vector[v]) myfile << 2;
			else if(Phase2Array[v] >= ContinuousArray[v] && Phase2Array[v] >= Phase1Array[v] && Phase2Array[v] >= Passivestd::vector[v]) myfile << 3;
*/
			// change "=" to random assignment? currently biases passive, then phase 1, then phase 2 if all equal

			myfile << ArrayForVoxelyze[v];
			md5file << ArrayForVoxelyze[v];

			v++;
		}
	
		//if (z<num_z_voxels-1) myfile << "]]></Layer><Layer><![";

	}
	if (z<num_z_voxels-1) {myfile << "]]></Layer>\n<Layer><![CDATA["; md5file << "\n";}
}


myfile << "]]></Layer>\n\
</Data>\n\
</Structure>\n\
</VXC>\n\
</VXA>";

  		myfile.close();
		md5file.close();
		
		int numVoxelsFilled = 0;
		int numVoxelsActuated = 0;
		
		// check if "robot" is empty space:
		for (int i=0; i<ArrayForVoxelyze.size(); i++)
		{
			if (!ArrayForVoxelyze[i]==0)
			{
				numVoxelsFilled++;
				if (ArrayForVoxelyze[i] == 1 || ArrayForVoxelyze[i] == 3)
				{
					numVoxelsActuated++;
				}
			}
		}
		//if (voxelPenalty > 1) {return true;}
		//else {return false;}
/*
		if (int(NEAT::Globals::getSingleton()->getParameterValue("PenaltyType")) == 3) 
		{
			return numConnections/2;
		} else {
			if (int(NEAT::Globals::getSingleton()->getParameterValue("PenaltyType")) == 2) 
			{
				return numVoxelsActuated;
			} else {
				return numVoxelsFilled;
			}
		}
*/
		return numVoxelsFilled;  //nac: hand coded Penalty Type
	}		


